package room

import (
	"fmt"
	"samba/pkg/log"
	"samba/stub"
	"samba/util/model"
	"samba/util/util"
)

type ItemLog struct {
	Id  int   `json:"id"`  // 物品id
	Num int64 `json:"num"` // 物品数量
}

type PlayerLog struct {
	UId           int64     `json:"uid"`           // 玩家id
	Seat          int       `json:"seat"`          // 座位号
	TakeCoin      int64     `json:"initial_coins"` // 携带金币
	WinCoin       int64     `json:"win_coins"`     // 输赢金币
	EndCoin       int64     `json:"end_coins"`     // 结束金币
	Tip           int64     `json:"charge"`        // 台费
	Raise         int       `json:"truco"`         // 叫分次数
	GiveUp        int       `json:"correrr"`       // 放弃次数
	ContinueRaise int       `json:"aumentar"`      // 继续加注次数
	Agree         int       `json:"aceitar"`       // 同意次数
	CostItem      []ItemLog `json:"cost_item"`     // 消耗物品
	Color         int       `json:"color"`         // 队伍颜色 0:红方 1:绿方
	Robot         int       `json:"robot"`         // 机器人标识 0:机器人 1：真人
	RobotTemper   int       `json:"robot_temper"`  // 机器人性格 	0:未知 1:稳健 2:激进 3:普通 4:摆烂 5:策略型
}

type TeamLog struct {
	Color    int `json:"color"`       // 队伍颜色
	Point    int `json:"play_points"` // 累积得分
	Deciding int `json:"player_key"`  // 决胜手次数
	BigPoker int `json:"player_king"` // 触发大牌次数
	Show     int `json:"player_show"` // 本局手牌展示次数
	Concert  int `json:"player_team"` // 行动一致次数
}

type TrucoGameLog struct {
	RoomId    int          `json:"room_id"`    // 房间id
	RoomType  int          `json:"room_type"`  // 房间类型
	Level     int          `json:"level"`      // 房间等级:低中高级场
	PlayType  int          `json:"play_type"`  // 玩法类型
	Blind     int64        `json:"bet"`        // 底注
	GameNo    string       `json:"game_no"`    // 本局唯一id
	StartTime int64        `json:"start_time"` // 开始时间
	EndTime   int64        `json:"end_time"`   // 结束时间
	Players   []*PlayerLog `json:"players"`    // 玩家数据
	Teams     []*TeamLog   `json:"teams"`      // 团队数据
	ClubId    int          `json:"club_id"`    // 当前俱乐部id 0:金币场
}

func NewTrucoGameLog(room *GameRoom) *TrucoGameLog {
	t := &TrucoGameLog{RoomId: room.Id(), Level: room.RoomCnf.Level, RoomType: room.Type(), PlayType: room.PlayType(), GameNo: room.gameNo,
		StartTime: room.startTime.Unix(), Blind: room.RoomCnf.Blind, ClubId: room.ClubId()}
	for _, seat := range room.Seats {
		player := seat.Player()
		robot := util.Tie(player.IsRobot(), 1, 0)
		temper := int(stub.RteUnknown)
		if player.IsRobot() {
			temper = int(player.Robot.Cnf())
		}
		t.Players = append(t.Players, &PlayerLog{UId: player.UID, Seat: seat.No(), TakeCoin: player.TakeCoin, Robot: robot, RobotTemper: temper})
	}
	return t
}

func (t *TrucoGameLog) Player(seatNo int) *PlayerLog {
	if seatNo < 0 || seatNo > len(t.Players) {
		log.Error(fmt.Sprintf("room:%v type:%v game log error.seat:%v is nil", t.RoomId, t.RoomType, seatNo))
		return &PlayerLog{}
	}
	return t.Players[seatNo]
}

func (t *TrucoGameLog) Flush(room *GameRoom) {
	t.EndTime = room.endTime.Unix()
	for _, seat := range room.Seats {
		if t.Player(seat.No()) == nil {
			log.Error(room.SeatLog(seat, "player is nil"))
		}
		t.Player(seat.No()).EndCoin = seat.Player().TakeCoin
		t.Player(seat.No()).WinCoin = t.Player(seat.No()).EndCoin - t.Player(seat.No()).TakeCoin
	}
	model.AddRoomGameLog(room.PlayType(), t)
}
